############################################################ 
# spectral clustering & scrubbing internals
############################################################ 

# .getLaplacian -- given S, return the Laplacian
# .getEigLapFromDist -- get spectral decomposition from distance matrix
# .getEigLapFromData -- get spectral decomposition from data
# .resamp.minSpect -- returns spectrum from resampled data
# .getSigEig -- compares true spectrum to resampled data, returning signigivant eigen values
# .getNClustsBIC -- estimates number of clusters using EM fitting of mixture models & choosing w. BIC
# .getClustersFromEigen -- given Laplacian eigenvecs & Nclusts, cluster
# .getScrubbed -- returns scrubbed data


####
# functions for computing the Laplacian & its eigendecomposition 
####

.getLaplacian <- function(S, norm=c("symmetric","rowsum","none")) {
	# Given S, return the Laplacian
  
	# Compute degree matrix
	D <- rowSums(S)
	
	L <- switch(norm[1],
		# symmetric normalized L
		## symmetric = I - (diag(D^(-1/2))) %*% S %*% (diag(D^(-1/2))), #slow
		symmetric = {sqrtD <- sqrt(D); (diag(D)-S)/outer(sqrtD,sqrtD)},
		# row-sum normalized L
		rowsum = {I <- diag(ncol(S)); I - diag(D^-1) %*% S},
		# unnormalized
		diag(D) - S
	)
	return(L)
}

.getEigLapFromDist <- function (d, sigma, norm, only.values=FALSE) {
	# given a distance matrix, get the eigenvals/vecs
	# of the Laplacian using method=scale.type in .getSimilMat
	# and method=lap.type in .getLaplacian

	N <- ncol(d)
	out <- eigen( 
		.getLaplacian( 
			S=.getSimilarity(d, sigma), 
			norm
		),
		only.values=only.values
	)
	out$values <- rev(out$values)
	out$vectors <- out$vectors[,N:1]
	return(out)
}

.getEigLapFromData <- function (x, distanceFn, sigma, norm, only.values=FALSE, ...) {
	# as above, but get it directly from the data.  other
	# function names besides "cord.dist" may be passed as needed
	# if some other metric is to be used
	# ... are additional parameters to be passed to the distfn 
	d <- do.call(distanceFn,list(x,...))
	.getEigLapFromDist(d, sigma, norm, only.values)
}

####
# functions for obtaining k & l 
####


# First order of business: is there really any structure present?
# How many eigenvalues from the true data are smaller than the 
# smallest Fiedler value obtained in resampled (bootstrap null) data?
# Expect at least two... 
# Resample the rows of the data (ie preserving the marginal row statistics,
# eg the yea/nea/abstain for each vote while resistributing who voted which way), 
# and fetch the smallest Fiedler value:

.resample.spectra <- function (x, runs, resample=c("distances","rows"), distanceFn, sigma, norm, ...) {
	# ... = additional parameters ultimately passed to distanceFn
	spectra <- switch(resample[1],
		rows = sapply (1:runs, function(run){
			# resample all rows without replacement, runs times
			# NB apply returns the data columnwise, so need to transpose back to rows
			x <- t(apply(x,1,sample))
			.getEigLapFromData(x, distanceFn, sigma, norm, only.values=TRUE, ...) 
		}),		
		distances = { 
			d = do.call(distanceFn,list(x,...))
			sapply (1:runs, function(run){
				# resample the distance values
				newd <- matrix(0,nr=nrow(d),nc=ncol(d))
				newd[upper.tri(newd)] <- sample(d[upper.tri(d)])
				newd <- newd+t(newd)
				diag(newd) <- diag(d)
				.getEigLapFromDist(newd, sigma, norm, only.values=TRUE)$values
			})
		}
	)
	out <- list(
		refGaps=apply(spectra[-1,]-spectra[-nrow(spectra),],1,quantile,0.95,1),
		refSpectrum=apply(spectra,1,quantile,0.05,1)
	)
	return(out)		
}


.getSignificantEigvals <- function(spectrum, nullSpect, compare=c("fiedgap","fiedval","gaps","vals")){
	Neig <- length(spectrum)
	gaps <- spectrum[-1]-spectrum[-Neig]
	out <- switch(compare[1],
		fiedgap = {
			sig <- (gaps>nullSpect$refGaps[2])&(gaps>nullSpect$refGaps)
			sig[1] <- TRUE # by defn the first one counts
			which(sig)
		},
		gaps = {
			sig <- gaps>nullSpect$refGaps
			sig[1] <- T # by defn the first one counts
			which(sig)
		},
		fiedval = {
			sig <- spectrum<nullSpect$refSpectrum[2]
			1:which(!sig[-1])[1]
		},
		vals = {
			sig <- spectrum<nullSpect$refSpectrum
			1:which(!sig[-1])[1]
		}
	)
	# often with fiedgap and gaps, we wind up with spurious "significant" ones
	# on the large eigenvalue end due to sampling; to protect against this
	# if there's a break in the run of more than three, truncate:
	out <- out[1:min(length(out),which((out[-1]-out[-length(out)])>3))]
	return(out)
}

# Second order of business: determine the number of clusters. 
# This is tricky; cf Still & Bialek 2004.  
# In roll call data paper, Pauls &al fit Gaussian
# mixture models to the distribution of the Fiedler
# vector, chosing the optimal number of components 
# using AIC. Here, we do similarly, but use BIC.  


.get.kClusts <- function (FiedlerVec, G=1:20) {
	
	# In roll call data paper, Pauls &al fit Gaussian
	# mixture models to the distribution of the Fiedler
	# vector, chosing the optimal number of components 
	# using AIC. Here, we do similarly, but use BIC.  

	# FiedlerVec is 2nd column from .getEigLap...()$vectors
	# G is the range of components to try; hard-coded up to 20
	# Mclust will attempt both equal and unequal variance mixtures
	# see the mclust library documentation for details
	# $G returns optimal G

	require (mclust)
	Mclust(FiedlerVec,G)$G
}

###
# Final step: scrubbing
###

.getScrubbed <- function(x,clusLabels,scrubOnProjectionFailure=FALSE){
	# return the scrubbed data by subtracting out the projection of the data
	# onto the centroids	
	
	# First take the output of .getClustersFromEigen and get the centroids
	centroids <- sapply(unique(clusLabels), function(label){
		rowMeans(x[,clusLabels==label,drop=FALSE],na.rm=TRUE)
	})

	# Compute LSQ projection of data onto centroids.  
	# (LSQ fitting is implemented in R using QR decompostition in linpack)
	# see ?lm for details of lsq fitting syntax.
	# ... note the 0 to kill off the intercept term and na.exclude to pad 
	#     where there are missing values.
	
	projection <- lm(x ~ 0 + centroids, na.action=na.exclude)
	
	# the output of lm contains the least squares fit as well as the
	# residuals; we'll use residuals() to extract it [scrubbing]
	
	# However, first check for  "projection failure" (lin dependent centroids)
	# We can scrub even if the centroids are linearly dependent; however, if they
	# are  linearly dependent, there may be interpretation and/or numeric stability
	# problems.  set fail.on.singular=TRUE to not give back the scrubbed data in this
	# case, but otherwise do it anyway.  Warns in either case.
		
	if (projection$rank!=ncol(centroids)) {
		if (!scrubOnProjectionFailure) {
			scrubbed <- matrix()
			cat("Projection Failure: cluster centroids are not independent. Stopping with current layer.\n")
		} else {
			warning("Projection Failure: cluster centroids are not independent. Returning scrubbed data, but proceed with care.")
			# the residuals function automatically does the subtraction
			scrubbed <- residuals(projection)
			colnames(scrubbed) <- colnames(x)
		}
	} else {
		scrubbed <- residuals(projection)
		colnames(scrubbed) <- colnames(x)
	}
	# t(scrubbed) #to get back to samples=rows
	return(scrubbed)
}







